/*
 * X11 GUI for slide puzzle game
 * Copyright © 2022-2023 Nick Bowler
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <assert.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>
#include <X11/Shell.h>

#include "motif.h"
#include "ewmhicon.h"

/* Size of the traditional icon pixmap (multiple of 3) */
#define ICON_SIZE 48

/* Define to 1 to add highlights and delays for visually debugging redraws */
#if X11_RENDER_DEBUG
#  include <unistd.h>
#endif

/* status flags in the upper bits are ORed into the game mask which only
 * needs the lower 25 bits */
#define X11_TICK_INSTALLED (1ul << 31)
#define X11_PROC_INSTALLED (1ul << 30)

enum {
	NUM_TILE_COLOURS = COLOUR_MAX*(TILE_MAX-1),
	COLOUR_UI_GAMEBG = NUM_TILE_COLOURS,
	COLOUR_UI_MAX
};

static void init_colours(struct app_state *state, Widget shell)
{
	static const char cdefs[] =
		     "colour0"        "\0" COLOUR0_PRIMARY
		"\0" "colourDark0"    "\0" COLOUR0_DARK
		"\0" "colourLight0"   "\0" COLOUR0_LIGHT
		"\0" "colour1"        "\0" COLOUR1_PRIMARY
		"\0" "colourDark1"    "\0" COLOUR1_DARK
		"\0" "colourLight1"   "\0" COLOUR1_LIGHT
		"\0" "colour2"        "\0" COLOUR2_PRIMARY
		"\0" "colourDark2"    "\0" COLOUR2_DARK
		"\0" "colourLight2"   "\0" COLOUR2_LIGHT
		"\0" "colour3"        "\0" COLOUR3_PRIMARY
		"\0" "colourDark3"    "\0" COLOUR3_DARK
		"\0" "colourLight3"   "\0" COLOUR3_LIGHT
		"\0" "colour4"        "\0" COLOUR4_PRIMARY
		"\0" "colourDark4"    "\0" COLOUR4_DARK
		"\0" "colourLight4"   "\0" COLOUR4_LIGHT
		"\0" "colour5"        "\0" COLOUR5_PRIMARY
		"\0" "colourDark5"    "\0" COLOUR5_DARK
		"\0" "colourLight5"   "\0" COLOUR5_LIGHT
		"\0" "gameBackground" "\0" "#181818";

	XtResource res[COLOUR_UI_MAX], t = {
		(char *)cdefs, 0, XtRPixel, sizeof (Pixel), 0, XtRString
	};
	Pixel colours[XtNumber(res)];
	unsigned i;

	for (i = 0; i < XtNumber(res); i++) {
		t.default_addr = strchr(t.resource_name, 0) + 1;
		res[i] = t;

		t.resource_name = (char *)t.default_addr + 8;
		t.resource_offset += sizeof colours[0];
	}

	XtGetApplicationResources(shell, &colours, res, XtNumber(res), NULL, 0);
	for (i = 0; i < NUM_TILE_COLOURS; i++) {
		uint_least32_t *p = state->tile_colour[0];
		p[i] = colours[i];
	}
	XtVaSetValues(state->game, XtNbackground, colours[i], (char *)NULL);
}

void x11_initialize(struct app_state *state, Widget shell)
{
	Display *display = XtDisplay(shell);
	Screen *screen = XtScreen(shell);
	Window root = RootWindowOfScreen(screen);
	XGCValues gcv;

	init_colours(state, shell);

	gcv.line_width = 1;
	state->tile_gc = XCreateGC(display, root, GCLineWidth, &gcv);

	state->icon_pixmap = XCreatePixmap(display, root, ICON_SIZE, ICON_SIZE,
	                                   DefaultDepthOfScreen(screen));
}

static void draw_tile(struct app_state *state, Display *display, Drawable d,
                      int tile, int gx, int gy, Dimension tw, Dimension th)
{
	int tx = gx * tw, ty = gy * th;

	XSegment topshadow[] = {
		{ tx,      ty,      tx,      ty+th   },
		{ tx+1,    ty,      tx+tw,   ty      },

		{ tx+1,    ty+1,    tx+1,    ty+th-1 },
		{ tx+2,    ty+1,    tx+tw-1, ty+1    }
	};

	XSegment bottomshadow[] = {
		{ tx+1,    ty+th-1, tx+tw,   ty+th-1 },
		{ tx+tw-1, ty+th-1, tx+tw-1, ty+1    },

		{ tx+2,    ty+th-2, tx+tw-1, ty+th-2 },
		{ tx+tw-2, ty+th-2, tx+tw-2, ty+2    }
	};

	XSetForeground(display, state->tile_gc, state->tile_colour[tile-1][COLOUR_LIGHT]);
	XDrawSegments(display, d, state->tile_gc, topshadow, XtNumber(topshadow));

	XSetForeground(display, state->tile_gc, state->tile_colour[tile-1][COLOUR_DARK]);
	XDrawSegments(display, d, state->tile_gc, bottomshadow, XtNumber(bottomshadow));

	XSetForeground(display, state->tile_gc, state->tile_colour[tile-1][COLOUR_PRIMARY]);
	XFillRectangle(display, d, state->tile_gc, tx+2, ty+2, tw-4, th-4);
}

static void
clear_rect(Display *display, Drawable d, int x, int y, int w, int h)
{
#if X11_RENDER_DEBUG
	XRectangle r = { x, y, w, h };
	XGCValues gcv;
	GC gc;

	gcv.foreground = 0xff0000;
	gc = XCreateGC(display, d, GCForeground, &gcv);
	XFillRectangles(display, d, gc, &r, 1);
	XFreeGC(display, gc);

	XFlush(display);
	usleep(70000);
#endif

	XClearArea(display, d, x, y, w, h, 0);
}

/*
 * Efficiently clear all the border tiles in the game area.  The mask indicates
 * which tiles need clearing, but for the border clear it is safe to wipe an
 * entire row or column of the border using a single XClearArea.
 */
static void clear_border(Display *display, Drawable d, unsigned sz)
{
	clear_rect(display, d, 0,    0,    -1, sz);
	clear_rect(display, d, 0,    4*sz, -1, sz);
	clear_rect(display, d, 0,    0,    sz, -1);
	clear_rect(display, d, 4*sz, 0,    sz, -1);
}

static int
redraw_tile(struct app_state *state, Display *display, Drawable d,
            uint_least32_t *gp, int x, int y, unsigned sz)
{
	unsigned tile = board_tile(gp, 5*y+x);
	unsigned w = sz, h = sz;

#if X11_RENDER_DEBUG
	if (d == XtWindow(state->game) || d == XtWindow(state->goal)) {
		XRectangle r = { x*w, y*h, w, h };
		XSetForeground(display, state->tile_gc, 0xff0000);
		XFillRectangles(display, d, state->tile_gc, &r, 1);
		XFlush(display);
		usleep(70000);
	}
#endif
	assert(tile < TILE_MAX);
	if (tile == TILE_EMPTY) {
		XClearArea(display, d, x*w, y*h, w, h, 0);
	} else {
		draw_tile(state, display, d, tile, x, y, w, h);
	}

	return tile;
}

static void set_icon(struct app_state *state, Display *display,
                     Widget shell, unsigned long *ewmhseq)
{
	/*
	 * Clear and reset XtNiconPixmap otherwise it seems dtwm will not
	 * notice the changed icon.
	 */
	XtVaSetValues(shell, XtNiconPixmap, None, (char *)NULL);
	XtVaSetValues(shell, XtNiconPixmap, state->icon_pixmap, (char *)NULL);

	if (state->flags & FLAG_USE_EWMH_ICONS) {
		Atom net_wm_icon = XInternAtom(display, "_NET_WM_ICON", FALSE);
		Colormap cmap = DefaultColormapOfScreen(XtScreen(shell));
		XColor colours[NUM_TILE_COLOURS];
		void *wm_icon;
		int i;

		for (i = 0; i < XtNumber(colours); i++) {
			uint_least32_t *p = state->tile_colour[0];
			colours[i].pixel = p[i];
		}

		XQueryColors(display, cmap, colours, XtNumber(colours));
		wm_icon = ewmh_icon_generate(ewmhseq, colours);
		if (wm_icon) {
			XChangeProperty(display, XtWindow(shell), net_wm_icon,
			                XA_CARDINAL, 32, PropModeReplace,
			                wm_icon, EWMH_ICON_NELEM);
		}
		free(wm_icon);
	}
}

void
x11_redraw_goal(struct app_state *state, uint_fast32_t mask, Widget icon_shell)
{
	Display *display = XtDisplay(state->goal);
	Window goal = XtWindow(state->goal);
	unsigned sz = state->goal_tile_sz;
	unsigned long ewmhseq[9];
	int i;

	uint_least32_t gp[3] = {
		state->board.goal[0],
		state->board.goal[1],
		state->board.goal[2]
	};

	for (i = 0; i < 9; i++) {
		int x = i%3, y = i/3;

		if (icon_shell) {
			unsigned tile;

			tile = redraw_tile(state, display, state->icon_pixmap,
			                   gp, x, y, ICON_SIZE/3);
			ewmhseq[i] = COLOUR_MAX*0x10101 * (tile-1) + 0x20100;
		}

		if (mask & 1)
			redraw_tile(state, display, goal, gp, x, y, sz);

		/*
		 * Goal bitmaps have a gap of 2 tiles between each row.
		 * This funny shift will accomodate that.
		 */
		mask >>= 1|x;
	}

	if (icon_shell)
		set_icon(state, display, icon_shell, ewmhseq);
}

void x11_redraw_game(struct app_state *state, uint_fast32_t mask)
{
	Display *display = XtDisplay(state->goal);
	Window game = XtWindow(state->game);
	uint_least32_t buf[3], *gp = state->board.game;
	unsigned sz = state->game_tile_sz;
	int i;

	if (state->flags & FLAG_VIEW_GOAL_ON_GAME)
		gp = game_overlay_goal(&state->board, buf);

	/* Optimize the game end case where the outer frame is cleared */
	if (mask & GAME_MASK & ~GOAL_MASK) {
		if (((gp[0] | gp[1] | gp[2]) & ~GOAL_MASK) == 0) {
			clear_border(display, game, sz);
			mask &= GOAL_MASK;
		}
	}

	for (i = 0; i < 25; i++) {
		if (mask & 1)
			redraw_tile(state, display, game, gp, i%5, i/5, sz);
		mask >>= 1;
	}
}

/*
 * Deferred redraw of tiles after resize/expose to avoid redundant drawing.
 *
 * Record any tiles that need to be redrawn due to resizes or expose events,
 * then, after a short delay, perform all the accumulated redraws at once.
 *
 * This is implemented using both a work proc and a timeout, because it seems
 * that rendering directly inside the timeout callback gives poor results:
 * possibly redrawing outdated intermediate positions during "fast" resizes
 * long after the resizing has stopped.  This does not happen when drawing
 * from a work proc.
 */
static Boolean do_render(void *data)
{
	struct app_state *state = data;

	x11_redraw_goal(state, state->render_goal_mask, NULL);
	x11_redraw_game(state, state->render_game_mask);

	state->render_goal_mask = state->render_game_mask = 0;
	return True;
}

static void start_render(void *data, XtIntervalId *id)
{
	struct app_state *state = data;
	XtAppContext app;

	if (state->render_game_mask & X11_PROC_INSTALLED)
		return;

	state->render_game_mask |= X11_PROC_INSTALLED;
	app = XtWidgetToApplicationContext(state->game);
	XtAppAddWorkProc(app, do_render, state);
}

void x11_queue_render(struct app_state *state, uint_fast32_t mask, int mode)
{
	uint_fast32_t changed = 0;
	XtAppContext app;

	mask &= GAME_MASK;
	if (mode & RENDER_MODE_GOAL)
		changed |= state->render_goal_mask |= mask;
	if (mode & RENDER_MODE_GAME)
		changed |= state->render_game_mask |= mask;

	if (changed & X11_TICK_INSTALLED || !changed)
		return;

	state->render_game_mask |= X11_TICK_INSTALLED;
	app = XtWidgetToApplicationContext(state->game);
	XtAppAddTimeOut(app, 3, start_render, state);
}
